﻿from geometry import Vector
from random import choice, random

class Entity:

    gravity = 30000
    drag = 25
    friction = 500
    coefficient_of_restitution = 0.4

    def __init__(self, image, position, velocity=Vector.zero()):
        self.image = image
        self.position = position
        self.velocity = velocity

    def update(self, time_step):
        self.position += self.velocity * time_step
        self.velocity.y -= self.gravity * time_step * time_step
        self.velocity -= self.velocity * (time_step * time_step * self.drag)

    def collision_box(self, width, height, center):
        sweeper = Vector(width/2, height/2)
        bl = self.position + Vector(20, 20) - sweeper
        tr = self.position + Vector(self.image.width, self.image.height) + sweeper - Vector(20, 20)
        return center.between(bl, tr)

    def points(self):
        w = self.image.width
        h = self.image.height
        x1, y1 = self.position.x, self.position.y
        x2, y2 = x1+w, y1+h
        return Vector(x1, y1), Vector(x2, y2)


class Barrier:

    max_life = 1000

    def __init__(self, x, y):
        self.rotation = random()*360
        self.rps = random()*360*2
        self.position = Vector(x, y)
        self.velocity = (random()-0.5)*0.5

    def update(self, time_step):
        pos = self.position
        self.position.x += self.velocity*time_step
        self.rotation += self.rps*time_step
        if self.position.x < 0.0 or self.position.x > 1.0:
            self.position = self.position
            self.velocity *= -1


class Bullet(Entity):

    max_life = 3.0
    gravity = 15000
    drag = 10

    def __init__(self, *args, **kwargs):
        self.die()
        super().__init__(*args, **kwargs)

    def update(self, time_step):
        if self.alive:
            super().update(time_step)
            self.life += time_step
            if self.life > self.max_life:
                self.die()

    def spawn(self, image, position, velocity):
        self.life = 0.0
        self.alive = True
        self.image = image
        self.position = position
        self.velocity = velocity

    def die(self, die=None):
        self.life = 0.0
        self.alive = False
        if die:
            die()

class Exhaust(Entity):

    max_life = 3.0
    burnout = 1.0
    drag = 100
    gravity = -1000

    def __init__(self, fire_image, smoke_image, *args, **kwargs):
        self.life = 0.0
        self.alive = False
        self.fire_image = fire_image
        self.smoke_image = smoke_image
        super().__init__(fire_image, *args, **kwargs)

    def update(self, time_step):
        if self.alive:
            super().update(time_step)
            self.life += time_step
            if self.life > self.max_life:
                alive = False
            elif self.life > self.burnout:
                self.image = self.smoke_image

    def spawn(self, position, velocity):
        self.life = 0.0
        self.alive = True
        self.image = self.fire_image
        self.position = position
        self.velocity = velocity


class Enemy(Entity):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.life = 10
        self.bottom_angle = random()*360.0
        self.top_angle = random()*360.0

    def update(self, time_step):
        self.bottom_angle += time_step*90.0
        self.top_angle -= time_step*60.0

    def respawn(self, position_x, position_y):
        self.position = Vector(position_x, position_y)

    def take_damage(self, amount, die):
        self.life -= amount
        if self.life <= 0:
            die()
            self.life = 10
            self.respawn(self.position.x+1500+random()*500, random()*720)


class Robot(Entity):

    seconds_per_shot = 0.15
    max_bullets = 50
    max_exhaust = 75
    bullet_speed = 600.0

    def __init__(self, image, position, bullet_images, fire_image, smoke_image, gun_image):
        super().__init__(image, position)
        self.aim = 0
        self.position = Vector(400, 400)
        self.seconds_per_shot = Robot.seconds_per_shot
        self.elapsed = 0.0
        self.bullet_images = bullet_images
        self.fire_image = fire_image
        self.smoke_image = smoke_image
        self.bullet_speed = Robot.bullet_speed
        self.bullets = [Bullet(self.bullet_images[0], Vector.zero()) for _ in range(self.max_bullets)]
        self.exhaust = [Exhaust(self.fire_image, self.smoke_image, Vector.zero()) for _ in range(self.max_exhaust)]
        self.gun_pivot = Vector(20, 16)
        self.next_exhaust = 0
        self.exhaust_timer = 0
        self.life = 3
        self.alive = True

    def update(self, time_step):
        super().update(time_step)
        self.elapsed += time_step

        for bullet in self.bullets:
            if bullet.alive:
                bullet.update(time_step)
            elif self.elapsed > self.seconds_per_shot:
                bullet.spawn(choice(self.bullet_images), self.nozzle.random(5), (self.velocity+Vector.from_angle_and_size(self.aim, self.bullet_speed)).random(100))
                self.elapsed -= self.seconds_per_shot

        for particle in self.exhaust:
            if particle.alive:
                particle.update(time_step)

        if self.exhaust_timer == 1:
            self.exhaust_timer = 0
            self.exhaust[self.next_exhaust].spawn(self.thruster.random(15), Vector(0, -200) + self.velocity.random(100))
            self.next_exhaust = (self.next_exhaust + 1) % self.max_exhaust
        else:
            self.exhaust_timer += 1

    @property
    def nozzle(self):
        return self.position + self.gun_pivot + Vector.from_angle_and_size(self.aim, 70.0) + Vector(-10, 30)

    @property
    def thruster(self):
        return self.position + Vector(29, 15)

    @property
    def shots_per_second(self):
        if self.seconds_per_shot <= 0.0:
            self.seconds_per_shot = Robot.seconds_per_shot
        return 1.0/self.seconds_per_shot

    @shots_per_second.setter
    def shots_per_second(self, shots_per_second):
        if shots_per_second > 0.0:
            self.seconds_per_shot = 1.0/shots_per_second
        else:
            self.seconds_per_shot = Robot.seconds_per_shot

    def handle_collisions(self, enemies, bullet_hit, enemy_die):
        my_center = self.position + Vector(self.image.width/2, self.image.height/2)
        for enemy in enemies:
            for bullet in self.bullets:
                if bullet.alive:
                    if enemy.collision_box(20, 20, bullet.position + Vector(10, 10)):
                        enemy.take_damage(1, enemy_die)
                        bullet.die(bullet_hit)
            if enemy.collision_box(self.image.width, self.image.height, my_center):
                enemy.take_damage(10, enemy_die)
                self.take_damage(1)

    def take_damage(self, amount):
        self.life -= amount
        if self.life <= 0:
            self.alive = False
