﻿import ctypes
import os
import time
from geometry import Vector
os.environ["PYSDL2_DLL_PATH"] = os.path.join(os.getcwd(),"bin")
import sdl2
from sdl2 import sdlimage
from sdl2 import sdlmixer
sdl2.SDL_Init(sdl2.SDL_INIT_EVERYTHING)
sdlimage.IMG_Init(sdlimage.IMG_INIT_PNG)
sdlmixer.Mix_OpenAudio(44100, sdlmixer.MIX_DEFAULT_FORMAT, 2, 1024)


class Image:

    def __init__(self, texture, width, height, base_x, base_y):
        self.texture = texture
        self.width = int(width)
        self.height = int(height)
        if base_x and base_y:
            self.base_vector = Vector(base_x, base_y)
            self.base = sdl2.SDL_Point(int(base_x), int(base_y))
        else:
            self.base_vector = Vector.zero()
            self.base = sdl2.SDL_Point(0, 0)

class Slider:

    def __init__(self, value, image, callback):
        self.value = value
        self.image = image
        self._callback = callback

    def callback(self, value):
        self.value = value
        self._callback(value)

class Platform:

    def __init__(self):
        self.resolution = (1280, 720)
        self.window = sdl2.SDL_CreateWindow(b"Data Wipe!", sdl2.SDL_WINDOWPOS_UNDEFINED, sdl2.SDL_WINDOWPOS_UNDEFINED, self.resolution[0], self.resolution[1], sdl2.SDL_WINDOW_SHOWN)
        self.renderer = sdl2.SDL_CreateRenderer(self.window, -1, sdl2.SDL_RENDERER_ACCELERATED | sdl2.SDL_RENDERER_PRESENTVSYNC)
        self.event = sdl2.SDL_Event()
        self.camera = Vector(0.0, 0.0)
        self.sliders = []
        self.dragging = None
        self.audio_channels = {"Master": -1,
                               "FX": 0,
                               "Music": 1}
        sdlmixer.Mix_AllocateChannels(2)

    def run(self):
        self.load()
        # prime the clock
        start_time = time.clock()
        while True:
            while sdl2.SDL_PollEvent(ctypes.byref(self.event)):
                if self.event.type == sdl2.SDL_QUIT:
                    return
                elif self.event.type == sdl2.SDL_MOUSEMOTION:
                    self.on_mouse_move(self.world_coordinates(self.event.motion.x, self.resolution[1] - self.event.motion.y))
                elif self.event.type == sdl2.SDL_MOUSEBUTTONDOWN:
                    button = self.event.button.button
                    x = self.event.button.x
                    y = self.event.button.y
                    if button == sdl2.SDL_BUTTON_LEFT:
                        if x < 200 and y < len(self.sliders*50):
                            # clicked on a slider
                            self.dragging = self.sliders[int(y/50)]
                        else:
                            self.on_left_click(self.world_coordinates(x, y))
                    elif button == sdl2.SDL_BUTTON_RIGHT:
                        self.on_right_click(self.world_coordinates(x, y))
                elif self.event.type == sdl2.SDL_MOUSEBUTTONUP:
                    button = self.event.button.button
                    x = self.event.button.x
                    y = self.event.button.y
                    if button == sdl2.SDL_BUTTON_LEFT:
                        if self.dragging:
                            if x > 200:
                                self.dragging.callback(1.0)
                            elif x < 0:
                                self.dragging.callback(0.0)
                            else:
                                self.dragging.callback(x/200.0)
                            self.dragging = None
                elif self.event.type == sdl2.SDL_KEYDOWN:
                    key = self.event.key.keysym.sym
                    if key == sdl2.SDLK_SPACE:
                        self.on_key_press("space")
                    else:
                        self.on_key_press("any")

            self.pre_render()
            self.render()
            self.post_render()

            # determine how long a frame took
            end_time = time.clock()
            time_step = end_time - start_time
            start_time = time.clock()

            # skip updating the game if a large pause happened
            # this is to alleviate physics glitches in times where python decides
            # to garbage collect for a long time
            if time_step < 3.0/60.0:
                self.update(time_step)

    def delay(self, milliseconds):
        sdl2.SDL_Delay(milliseconds)

    def world_coordinates(self, x, y):
        return Vector(x, y) + self.camera

    def load(self):
        raise NotImplementedError("load method should be implemented entirely in the subclass")

    def load_image(self, filename, width, height, base_x, base_y):
        surface = sdlimage.IMG_Load(filename.encode("UTF-8"))
        texture = sdl2.SDL_CreateTextureFromSurface(self.renderer, surface)
        sdl2.SDL_FreeSurface(surface)
        image = Image(texture, width, height, base_x, base_y)
        return image

    def load_sound(self, filename):
        return sdlmixer.Mix_LoadWAV(filename.encode('UTF-8'))

    def set_channel_volume(self, channel, volume):
        assert volume >= 0 and volume <= 1
        sdlmixer.Mix_Volume(self.audio_channels[channel], int(volume*sdlmixer.MIX_MAX_VOLUME))

    def play_sound(self, sound, channel, loop=False, repeat=1, volume=1.0):
        sdlmixer.Mix_PlayChannel(self.audio_channels[channel], sound, -1 if loop else repeat - 1)

    def update(self, time_step):
        raise NotImplementedError("update method should be implemeneted entirely in the subclass")

    def render_image(self, image, position, rotation=0, paralax=1.0):
        pos = position - (self.camera * paralax)
        rect = sdl2.SDL_Rect(int(pos.x), int(self.resolution[1] - pos.y - image.height), image.width, image.height)
        sdl2.SDL_RenderCopyEx(self.renderer, image.texture, None, rect, -rotation, image.base, 0)

    def pre_render(self):
        sdl2.SDL_SetRenderDrawColor(self.renderer, 12, 17, 20, 255)
        sdl2.SDL_RenderClear(self.renderer)

    def render(self):
        raise NotImplementedError("render method should be implemented entirely in the subclass")

    def add_slider(self, image, slider):
        self.sliders.append(Slider(0.5, image, slider))

    def post_render(self):
        i = 0
        sdl2.SDL_SetRenderDrawColor(self.renderer, 0, 127, 0, 255)
        for slider in self.sliders:
            outer_rect = sdl2.SDL_Rect(0, i*50, 200, 50)
            inner_rect = sdl2.SDL_Rect(0, i*50+4, int(slider.value*200), 42)
            sdl2.SDL_RenderFillRect(self.renderer, inner_rect)
            sdl2.SDL_RenderCopy(self.renderer, slider.image.texture, None, outer_rect)
            i+=1

        sdl2.SDL_RenderPresent(self.renderer)

    def bounding_box(self, point1, point2):
        p1 = point1-self.camera
        p2 = point2-self.camera

        p1.y = self.resolution[1]-p1.y
        p2.y = self.resolution[1]-p2.y

        if p1.x < p2.x:
            x1 = p1.x
            x2 = p2.x
        else:
            x1 = p2.x
            x2 = p1.x

        if p1.y < p2.y:
            y1 = p1.y
            y2 = p2.y
        else:
            y1 = p2.y
            y2 = p1.y

        return sdl2.SDL_Rect(int(x1), int(y1), int(x2-x1), int(y2-y1))

    def display_bounding_box(self, entity):
        rect = self.bounding_box(*entity.points())
        sdl2.SDL_SetRenderDrawColor(self.renderer, 255, 0, 0, 255)
        sdl2.SDL_RenderDrawRect(self.renderer, rect)

    def on_left_click(self, coordinates):
        pass

    def on_right_click(self, coordinates):
        pass

    def on_mouse_move(self, coordinates):
        pass

    def on_key_press(self, key):
        pass
